#ifndef CONFIG_H
#define CONFIG_H

#include <limits.h>
#include <stdbool.h>
#include <X11/Xlib.h>
#include <X11/X.h>
#include <sys/types.h>

#define CONFIG_COMMAND_MAX_LEN 128
#define CONFIG_PARSERRESULT_MAX 1024
#define CONFIG_ARGC_MAX 16
#define CONFIG_ARGLEN_MAX 256

#define PARSER_RETURN(result, _success, fmt, ...)        \
    do { (result).success = (_success);                  \
    snprintf((result).message, CONFIG_PARSERRESULT_MAX, \
    (fmt), ##__VA_ARGS__); return (result);} while (0);

typedef struct Client Client;
typedef struct Wm Wm;
typedef struct Arg Arg;
typedef struct Keybind Keybind;

struct Arg {
    int i;
    unsigned int ui;
    float f;
    Client *c;
    char* s;
    char **sl;

    int count;
};

struct Keybind {
    unsigned int mask;
    KeySym keysym;
    void (*func)(Wm*, Arg*);
    Arg args;
};

typedef enum {
    TYPE_INTEGER,
    TYPE_UNSIGNED_INTEGER,
    TYPE_BOOL,
    TYPE_STRING,
    TYPE_KEYBIND,
    TYPE_FLOAT,
} CommandArgType;

typedef struct {
    bool success;
    char message[CONFIG_PARSERRESULT_MAX];
} ParserResult;

typedef struct {
    char string[CONFIG_COMMAND_MAX_LEN];
    size_t arg_count;
    off_t cfg_var_offset[CONFIG_ARGC_MAX];
    CommandArgType arg_type;
    size_t argc;
    char **argv;
    Arg *args;
} ConfigCommand;

typedef struct {
    char file_path[PATH_MAX];
    ConfigCommand *available_commands;
    size_t command_count;
} ConfigFile;

typedef struct {
    float ms_p;
    char *border_col;
    char *focused_border_col;
    unsigned int border_width;
    Keybind *keybinds;
    int kb_count;
    bool focus_on_motion;
    char log_path[PATH_MAX];
} Config;

void wm_cfg_init_def(Config *config);
void wm_keybinds_init_def(Config *config);
void wm_configfile_init(ConfigFile *config, const char *path);
void wm_configfile_free(ConfigFile *config);
void wm_configfile_read(ConfigFile *cfile, Config *config);
void wm_config_apply_command(Config *config, ConfigCommand *command);
ParserResult wm_configfile_parse_line(const ConfigFile *cfile, Config *config,
                                      char* line, ConfigCommand *command);
void wm_configcommand_init(ConfigCommand *command);
void wm_configcommand_copy(ConfigCommand *dest, const ConfigCommand *src);
void wm_configcommand_argv_parse(ConfigCommand *command);
void wm_configcommand_free(ConfigCommand *command);
bool atobool(char *arr);
Keybind wm_config_parse_bind_command(ConfigCommand *command);

#endif // CONFIG_H
